Require Import Relations RelationClasses.

From CoMoZFC Require Import Axioms.

Definition morph (x y : S) := {m : S | elem m x} -> {m : S | elem m y}.

Definition comp {x y z} (g : morph y z) (f : morph x y) : morph x z.
Proof.
  intros m.
  apply g.
  apply f.
  exact m.
Defined.

Example id (x : S) : morph x x.
Proof.
  intro; assumption.
Defined.

Definition bimorph (x y z : S) := {m : S | elem m x} -> {m : S | elem m y} -> {m : S | elem m z}.

Definition bijective {x y} (f : morph x y) : Prop := 
  exists (g : morph y x), 
    (
      forall a, 
        g (f a) = a
    ) /\
    forall a,
      f (g a) = a.

Example id_bijective {x} : bijective (id x).
Proof.
  exists (id x);
  firstorder.
Qed.

Inductive setsizeeq : relation S :=
  | setsizeeq_1 : forall x y, (exists (f : morph x y), bijective f) -> setsizeeq x y.

Instance setsizeeq_reflexive : Reflexive setsizeeq.
Proof.
  intro.
  constructor.
  eexists.
  exact id_bijective.
Qed.

Instance setsizeeq_symmetric : Symmetric setsizeeq.
Proof.
  intros ? ? [? ? [f [g []]]];
  constructor;
  exists g;
  exists f;
  split; assumption.
Qed.

Instance setsizeeq_transitive : Transitive setsizeeq.
Proof.
  intros ? ? ? H1 H3.
  destruct H1 as [? b [f [g [H1 H2]]]].
  remember b.
  destruct H3 as [? ? [h [i [H3 H4]]]]; subst.
  constructor.
  exists (comp h f).
  exists (comp g i).
  split.
  +
    intro.
    unfold comp.
    rewrite H3.
    rewrite H1.
    reflexivity.
  +
    intro.
    unfold comp.
    rewrite H2.
    rewrite H4.
    reflexivity.
Qed.

Instance setsizeeq_equivalence : Equivalence setsizeeq.
Proof.
  split.
  -
    exact setsizeeq_reflexive.
  -
    exact setsizeeq_symmetric.
  -
    exact setsizeeq_transitive.
Qed.
